module util.macro;

private {
    import util.str;

    import std.stdio;

}

struct Substitute
{
    char[] Pattern;
    char[] Resolution;
    char[] EscapeOpen;
    char[] EscapeClose;
}

struct TextSpan
{
    char[] Text;
    int Start;
    int End;
}

void append(inout Substitute[] pList, char[] pPattern, char[] pResolution)
{
    pList.length = pList.length + 1;
    pList[$-1].Pattern = pPattern;
    pList[$-1].Resolution = pResolution;
}

TextSpan findspan(char[] pText, char[] pToken, uint lNextScan)
{
    return *(new TextSpan);
}

TextSpan matchspan(char[] pText, char[] pToken, uint lNextScan)
{
    return *(new TextSpan);
}

bool RunMacro( inout char[] pText, in Substitute[] pMacros)
{
    bool lChanged;

    lChanged = false;

    // Check that this is really a text buffer and not a binary buffer.
    for(int i = 0; i < 1000 && i < pText.length; i++)
    {
        if (pText[i] > 127) // Sorry, ASCII only for now.
            return false;

        if (pText[i] < 8)
            return false;

        if ((pText[i] > 13) && (pText[i] < ' '))
            return false;
    }

    foreach(Substitute lMacro; pMacros)
    {
        TextSpan lPos;
        uint lNextPos;
        char[][] lPatternTokens;
        int lStartPos;
        int lEndPos;
        int lTokenIndex;
        uint lNextScan;
        bool lNeedsChanging;
        TextSpan[] lSpans;

        // Split the pattern into separate words.
        lPatternTokens = std.string.split(lMacro.Pattern, " ");
        if (lPatternTokens.length == 0)
            continue;

        // Now find the set of adjacent words in the text.
        lNextPos = 0;
        lNextScan = 0;
        lNeedsChanging = false;
        lStartPos = -1;

        while(true)
        {
            // Locate start of pattern by scanning through until
            // a match for the first token is found.
            lTokenIndex = 0;
            lPos = findspan(pText, lPatternTokens[lTokenIndex], lNextScan);
            if (lPos.Start == -1)
            {
                // The first token in the pattern was not found,
                // so move on to the next pattern.
                break;
            }

            // Record where to start looking for the next matching token.
            lNextScan = lPos.End;
            /* This is a valid match only if ...
                ** It is at the start of the line, or
                ** preceded by a whitespace, or
                ** starts with an alpha and preceded by a non-alpha, or
                ** starts with a non-alpha and preceded by an alpha
            */
            if ((lPos.Start != 0) &&
                (std.ctype.isspace(pText[lPos.Start-1]) == 0) &&
                ((std.ctype.isalnum(pText[lPos.Start]) != 0) ==
                 (std.ctype.isalnum(pText[lPos.Start-1]) != 0)
                 ))
                // Not a valid token, so scan the text beyond where
                // this bad token was found, looking for a good one.
                continue;

            // Record the token's starting point in the text.
            lStartPos = lPos.Start;

            // Locate the rest of the words.
            while(lTokenIndex < lPatternTokens.length)
            {
                // Record where to continue searching from.
                lNextPos = lPos.End;

                if (lNextPos >= pText.length)
                    // Ran out of text, so stop looking for tokens.
                    break;

                if (std.ctype.isspace(pText[lNextPos]) == 0)
                {
                    lNeedsChanging = false;
                    break;
                }
                // Save potential last position.
                lEndPos = lNextPos;

                // Skip over any spaces
                while( (lNextPos < pText.length) && (std.ctype.isspace(pText[lNextPos])))
                    lNextPos++;

                // Move on to the next pattern token.
                lTokenIndex++;
                if (lTokenIndex >= lPatternTokens.length)
                {   // No more tokens left.
                    if (lEndPos >= pText.length)
                    {
                        lNeedsChanging = true;
                        break;
                    }

                    if (std.ctype.isalnum(pText[lEndPos-1]) != 0)
                    {
                        if (std.ctype.isalnum(pText[lEndPos]) == 0)
                        {
                            // The last char of the last token is an alnum,
                            // so the next character must not be.
                            lNeedsChanging = true;
                        }
                        else if (std.ctype.isspace(pText[lEndPos]) != 0)
                        {
                            // The last char of the last token is not an alnum,
                            // so the next character must be.
                            lNeedsChanging = true;
                        }
                    }
                    else
                    {
                        if (std.ctype.isalnum(pText[lEndPos]) != 0)
                        {
                            // The last char of the last token is an alnum,
                            // so the next character must not be.
                            lNeedsChanging = true;
                        }
                        else if (std.ctype.isspace(pText[lEndPos]) != 0)
                        {
                            // The last char of the last token is not an alnum,
                            // so the next character must be.
                            lNeedsChanging = true;
                        }
                    }

                    break;
                }

                lPos = matchspan(pText, lPatternTokens[lTokenIndex], lNextScan);
                if (lPos.Start == -1 )
                {
                    lNeedsChanging = false;
                    break;
                }

                if (lPos.Start >= pText.length)
                {
                    lNeedsChanging = false;
                    break;
                }

                lEndPos = lPos.End;
                if (lEndPos >= pText.length)
                {
                    lNeedsChanging = false;
                    break;
                }

            }
            // Resolve the pattern if found.
            if (lNeedsChanging)
            {
                char[] lResolved;
                char[] lSpecial;

                lResolved = lMacro.Resolution.dup;

                // Deal with any forced space markers.
                lSpecial = lMacro.EscapeOpen ~ "s" ~ lMacro.EscapeClose;
                if (lResolved.begins(lSpecial))
                    if (std.ctype.isspace(pText[lStartPos-1]))
                        lResolved = lResolved[lSpecial.length..$];
                    else
                        lResolved = " " ~ lResolved[lSpecial.length..$];


                if (lResolved.ends(lSpecial))
                    if (std.ctype.isspace(pText[lEndPos]))
                        lResolved = lResolved[0..$-lSpecial.length];
                    else
                        lResolved = lResolved[0..$-lSpecial.length] ~ " ";

                // replace pattern with resolution.
                pText = pText[0 .. lStartPos] ~
                        lResolved ~
                        pText[lEndPos .. $];
                lChanged = true;

                // Calculate next point to resume scanning.
                lNextScan = lStartPos + lResolved.length;
            }
            }
    }

    return lChanged;
}
