/**************************************************************************

        @file str.d

        Copyright (c) 2005 Derek Parnell

        This software is provided 'as-is', without any express or implied
        warranty. In no event will the authors be held liable for damages
        of any kind arising from the use of this software.

        Permission is hereby granted to anyone to use this software for any
        purpose, including commercial applications, and to alter it and/or
        redistribute it freely, subject to the following restrictions:

        1. The origin of this software must not be misrepresented; you must
           not claim that you wrote the original software. If you use this
           software in a product, an acknowledgment within documentation of
           said product would be appreciated but is not required.

        2. Altered source versions must be plainly marked as such, and must
           not be misrepresented as being the original software.

        3. This notice may not be removed or altered from any distribution
           of the source.

        4. Derivative works are permitted, but they must carry this notice
           in full and credit the original source.


                        ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


        @version        Initial version, January 2005
        @author         Derek Parnell


**************************************************************************/

///topic Strings
///info
// Miscellaneous routines and symbols used when working with character strings

module util.str;
private import util.str_bn;

private{
    import std.utf;
    import std.ctype;
    import util.linetoken;

    import std.string;
    import std.stdio;
}

public{

    debug(1)
    {
        private import std.stdio;
        static this()
        {
        writefln("util.str build #%d", auto_build_number);
        }
    }


///topic Strings
///proc ReplaceChar(inout stringtype pText, dchar pFrom, dchar pTo)
///desc Replace all occurances a character with another.
//This examines each character in /i pText and where it matches /i pFrom
// it is replaced with /i pTo. /n
///b Note that /i pText can be a /char[], /wchar[], /dchar[] or /ubyte[] datatypes. /n
///b Note that if /i pText is a ubyte[] type, the /i pFrom and /i pTo must
//be either /b ubyte or /b char datatypes.
//
//Example:
///code
//  char[] test;
//  test = "abc,de,frgh,ijk,kmn";
//  ReplaceChar( test, ',', ':');
//  assert(test == "abc:de:frgh:ijk:kmn");
///endcode

//----------------------------------------------------------
ubyte[] ReplaceChar(in ubyte[] pText, ubyte pFrom, ubyte pTo)
//----------------------------------------------------------
out (pResult){
    ubyte[] lTempA;
    ubyte[] lTempB;

    assert( ! (pResult is null) );

    lTempA = pText;
    lTempB = pResult;
    assert(lTempA.length == lTempB.length);
    if ( pFrom != pTo)
    {
        foreach (uint i, ubyte c; lTempA)
        {
            if (c == pFrom)
            {
                assert(pTo == lTempB[i]);
            }
            else
            {
                assert(lTempA[i] == lTempB[i]);
            }
        }
    }
    else
        assert(lTempA == lTempB);
}
body {

    if (pFrom == pTo)
        return pText;
    else {
    foreach( uint i, inout ubyte lTestChar; pText){
        if(lTestChar == pFrom) {
            ubyte[] lTemp = pText.dup;
            foreach( inout ubyte lTestChar; lTemp[i .. length]){
                if(lTestChar == pFrom) {
                    lTestChar = pTo;
                }
            }

            return lTemp;
        }
        else {
            if (i == pText.length-1)
            {
                return pText;
            }
            else
                continue;
        }

    }
}
    assert(0);
}

//----------------------------------------------------------
ubyte[] ReplaceChar(in ubyte[] pText, char pFrom, char pTo)
//----------------------------------------------------------
{
    return ReplaceChar( pText, cast(ubyte)pFrom, cast(ubyte)pTo);
}


//----------------------------------------------------------
dchar[] ReplaceChar(in dchar[] pText, dchar pFrom, dchar pTo)
//----------------------------------------------------------
out (pResult){
    dchar[] lTempA;
    dchar[] lTempB;

    assert( ! (pResult is null) );

    lTempA = pText;
    lTempB = pResult;
    assert(lTempA.length == lTempB.length);
    if ( pFrom != pTo)
    {
        foreach (uint i, dchar c; lTempA)
        {
            if (c == pFrom)
            {
                assert(pTo == lTempB[i]);
            }
            else
            {
                assert(lTempA[i] == lTempB[i]);
            }
        }
    }
    else
        assert(lTempA == lTempB);
}
body {
    if (pFrom == pTo)
        return pText;

    else {
    foreach( uint i, inout dchar lTestChar; pText){
        if(lTestChar == pFrom) {
            dchar[] lTemp = pText.dup;
            foreach( inout dchar lTestChar; lTemp[i .. length]){
                if(lTestChar == pFrom) {
                    lTestChar = pTo;
                }
            }

            return lTemp;
        }


    }
    // If I get here, no changes were done.
    return pText;
}
}


//----------------------------------------------------------
char[] ReplaceChar(in char[] pText, dchar pFrom, dchar pTo)
//----------------------------------------------------------
body {

    return toUTF8( ReplaceChar( toUTF32(pText), pFrom, pTo) );

}

//----------------------------------------------------------
wchar[] ReplaceChar(in wchar[] pText, dchar pFrom, dchar pTo)
//----------------------------------------------------------
body {

    return toUTF16( ReplaceChar( toUTF32(pText), pFrom, pTo) );

}

//----------------------------------------------------------
ubyte[] toStringZ(ubyte[] pData)
//----------------------------------------------------------
{
    ubyte[] lTemp;

    lTemp = pData.dup;
    lTemp ~= '\0';
    return lTemp;
}

//----------------------------------------------------------
char[] toStringZ(char[] pData)
//----------------------------------------------------------
{
    char[] lTemp;

    lTemp = pData.dup;
    lTemp ~= '\0';
    return lTemp;
}

//----------------------------------------------------------
ubyte[] toStringA(char[] pData)
//----------------------------------------------------------
{
    ubyte[] lTemp;

    lTemp.length = pData.length;
    foreach( int i, char lCurrChar; pData)
    {
        lTemp[i] = cast(ubyte)lCurrChar;
    }

    return lTemp;
}
//----------------------------------------------------------
ubyte[] toStringA(wchar[] pData)
//----------------------------------------------------------
{
    ubyte[] lTemp;
    char[] lInter;

    lInter = toUTF8( pData );
    lTemp.length = lInter.length;
    foreach( int i, char lCurrChar; lInter)
    {
        lTemp[i] = cast(ubyte)lCurrChar;
    }

    return lTemp;
}
//----------------------------------------------------------
ubyte[] toStringA(dchar[] pData)
//----------------------------------------------------------
{
    ubyte[] lTemp;
    char[] lInter;

    lInter = toUTF8( pData );
    lTemp.length = lInter.length;
    foreach( int i, char lCurrChar; lInter)
    {
        lTemp[i] = cast(ubyte)lCurrChar;
    }

    return lTemp;
}
}

//----------------------------------------------------------
//----------------------------------------------------------
//----------------------------------------------------------
debug(1) private import std.stdio;
unittest{
    debug(2) writefln("%s", "str.UT01: b = ReplaceChar ( char[] a, lit, lit) ");
    char[] testA;
    char[] testB;
    char[] testC;

    testA = "abcdbefbq";
    testB = "a cd ef q";
    testC = ReplaceChar(testA, 'b', ' ');
    assert( testC == testB);
    assert( testA != testB);
}

unittest{
    debug(2) writefln("%s", "str.UT02: a = ReplaceChar ( char[] a, lit, lit) ");
    char[] testA;
    char[] testB;
    char[] testC;

    testA = "abcdbefbq";
    testB = "a cd ef q";
    testA = ReplaceChar(testA, 'b', ' ');
    assert( testA == testB);
}

unittest{
    debug(2) writefln("%s", "str.UT03: ReplaceChar ( char[], 'b', 'b') ");
    char[] testA;
    char[] testB;
    char[] testC;

    testA = "abcdbefbq";
    testB = "abcdbefbq";
    testC = ReplaceChar(testA, 'b', 'b');
    assert( testC == testB);
    assert( testA == testB);
}

unittest{
    debug(2) writefln("%s", "str.UT04: ReplaceChar ( wchar[], lit, lit) ");
    wchar[] testA;
    wchar[] testB;
    wchar[] testC;

    testA = "abcdbefbq";
    testB = "a cd ef q";
    testC = ReplaceChar(testA, 'b', ' ');
    assert( testC == testB);
    assert( testA != testB);
}

unittest{
    debug(2) writefln("%s", "str.UT05: ReplaceChar ( dchar[], lit, lit) ");
    dchar[] testA;
    dchar[] testB;
    dchar[] testC;

    testA = "abcdbefbq";
    testB = "a cd ef q";
    testC = ReplaceChar(testA, 'b', ' ');
    assert( testC == testB);
    assert( testA != testB);
}

unittest{
    debug(2) writefln("%s", "str.UT06: ReplaceChar ( char[], wchar, wchar) ");
    char[] testA;
    char[] testB;
    char[] testC;
    wchar f,t;

    testA = "abcdbefbq";
    testB = "a cd ef q";
    f = 'b';
    t = ' ';
    testC = ReplaceChar(testA, f, t);
    assert( testC == testB);
    assert( testA != testB);

}

unittest{
    debug(2) writefln("%s", "str.UT07: ReplaceChar ( dchar[], dchar, dchar) ");
    dchar[] testA;
    dchar[] testB;
    dchar[] testC;
    dchar f,t;

    testA = "abcdbefbq";
    testB = "axcdxefxq";
    f = 'b';
    t = 'x';
    testC = ReplaceChar(testA, f, t);
    assert( testC == testB);
    assert( testA != testB);
}

unittest{
    debug(2) writefln("%s", "str.UT08: ReplaceChar ( ubyte[], ubyte, ubyte) ");
    ubyte[] testA;
    ubyte[] testB;
    ubyte[] testC;
    ubyte f,t;

    testA = toStringA(cast(char[])"abcdbefbq");
    testB = toStringA(cast(char[])"a cd ef q");
    f = 'b';
    t = ' ';
    testC = ReplaceChar(testA, f, t);
    assert( testC == testB);
    assert( testA != testB);
}

unittest{
    debug(2) writefln("%s", "str.UT09: ReplaceChar ( ubyte[], char, char) ");
    ubyte[] testA;
    ubyte[] testB;
    ubyte[] testC;
    char f,t;

    testA = toStringA(cast(char[])"abcdbefbq");
    testB = toStringA(cast(char[])"a cd ef q");
    f = 'b';
    t = ' ';
    testC = ReplaceChar(testA, f, t);
    assert( testC == testB);
    assert( testA != testB);
}

    // ---- toStringA() -----
unittest{
    debug(2) writefln("%s", "str.UT10: toStringA( char[]) ");
    ubyte[] testA;
    char[] testB;
    ubyte[] testC;
    int c;

    testB = "abcdef";
    testA = toStringA(testB);
    c = 0;
    assert(testA.length == testB.length);
    foreach (uint i, ubyte x; testA)
    {
        if (x == testB[i])
            c++;
    }
    assert( c == testA.length);
}

unittest{
    debug(2) writefln("%s", "str.UT11: toStringA( wchar[]) ");
    ubyte[] testA;
    wchar[] testB;
    ubyte[] testC;
    int c;

    testB = "abcdef";
    testA = toStringA(testB);
    c = 0;
    assert(testA.length == testB.length);
    foreach (uint i, ubyte x; testA)
    {
        if (x == testB[i])
            c++;
    }
    assert( c == testA.length);
}

unittest{
    debug(2) writefln("%s", "str.UT12: toStringA( dchar[]) ");
    ubyte[] testA;
    dchar[] testB;
    ubyte[] testC;
    int c;

    testB = "abcdef";
    testA = toStringA(testB);
    c = 0;
    assert(testA.length == testB.length);
    foreach (uint i, ubyte x; testA)
    {
        if (x == testB[i])
            c++;
    }
    assert( c == testA.length);
}

public {
/*
    IsLike does a simple pattern matching process. The pattern
    string can contain special tokens ...
       '*'   Represents zero or more characters in the text.
       '?'   Represents exactly one character in the text.
       '\'   Is the escape pattern. The next text character
             must exactly match the character following the
             escape character. This is used to match the
             tokens if they appear in the text.


*/
bool IsLike(dchar[] pText, dchar[] pPattern)
body {
    const dchar kZeroOrMore = '*';
    const dchar kExactlyOne = '?';
    const dchar kEscape     = '\\';
    uint  lTextIndex = 0;
    uint  lPatternIndex = 0;

    while( (lTextIndex < pText.length) || (lPatternIndex < pPattern.length) )
    {
        if ( (lPatternIndex < pPattern.length) &&
             (pPattern[lPatternIndex] == kZeroOrMore)
           ) {
            do lPatternIndex++;
                while( (lPatternIndex < pPattern.length) &&
                       (pPattern[lPatternIndex] == kZeroOrMore)
                     );
            if (lPatternIndex == pPattern.length)
                return true;
            else {
            if (pPattern[lPatternIndex] != kExactlyOne)
                while( (lTextIndex < pText.length) &&
                         ((pText[lTextIndex] != pPattern[lPatternIndex]) ||
                          (pPattern[lPatternIndex] == kExactlyOne)
                         )
                        )
                    lTextIndex++;
                }
        }
        else if ((lTextIndex < pText.length) &&
                 (lPatternIndex < pPattern.length)
                )
        {
            if ( (pPattern[lPatternIndex] == kEscape) &&
                 (lPatternIndex < pPattern.length - 1)
               )
                lPatternIndex++;

            if ( (pText[lTextIndex] == pPattern[lPatternIndex]) || (pPattern[lPatternIndex] == kExactlyOne) ) {
                lPatternIndex++;
                lTextIndex++;
            }
            else
            {
                return false;
            }
        }
        else
            return false;
    }

    return true;
}

bool IsLike(char[] pText, dchar[] pPattern)
{
    return IsLike( toUTF32(pText), pPattern);
}

bool IsLike(wchar[] pText, dchar[] pPattern)
{
    return IsLike( toUTF32(pText), pPattern);
}

bool IsLike(char[] pText, char[] pPattern)
{
    return IsLike( toUTF32(pText), toUTF32(pPattern));
}

bool IsLike(wchar[] pText, char[] pPattern)
{
    return IsLike( toUTF32(pText), toUTF32(pPattern));
}

bool IsLike(dchar[] pText, char[] pPattern)
{
    return IsLike( pText, toUTF32(pPattern));
}

bool IsLike(char[] pText, wchar[] pPattern)
{
    return IsLike( toUTF32(pText), toUTF32(pPattern));
}

bool IsLike(wchar[] pText, wchar[] pPattern)
{
    return IsLike( toUTF32(pText), toUTF32(pPattern));
}

bool IsLike(dchar[] pText, wchar[] pPattern)
{
    return IsLike( pText, toUTF32(pPattern));
}

}


//-------------------------------------------------------
int begins(char[] pString, char[] pSubString)
//-------------------------------------------------------
{
    int lResult = begins( std.utf.toUTF32(pString), std.utf.toUTF32(pSubString));

    if (lResult >= 0)
        lResult = std.utf.toUTFindex(pString, lResult);
    return lResult;
}

//-------------------------------------------------------
int begins(wchar[] pString, wchar[] pSubString)
//-------------------------------------------------------
{
    int lResult = begins( std.utf.toUTF32(pString), std.utf.toUTF32(pSubString));

    if (lResult >= 0)
        lResult = std.utf.toUTFindex(pString, lResult);
    return lResult;
}

//-------------------------------------------------------
int begins(dchar[] pString, dchar[] pSubString)
//-------------------------------------------------------
{
    if (pString.length < pSubString.length)
        return -1;
    if (pSubString.length == 0)
        return -1;

    if (pString[0..pSubString.length] == pSubString)
        return pSubString.length;
    return -1;
}

//-------------------------------------------------------
int ends(char[] pString, char[] pSubString)
//-------------------------------------------------------
{
    int lResult = ends( std.utf.toUTF32(pString), std.utf.toUTF32(pSubString));

    if (lResult >= 0)
        lResult = std.utf.toUTFindex(pString, lResult);
    return lResult;
}

//-------------------------------------------------------
int ends(wchar[] pString, wchar[] pSubString)
//-------------------------------------------------------
{
    int lResult = ends( std.utf.toUTF32(pString), std.utf.toUTF32(pSubString));

    if (lResult >= 0)
        lResult = std.utf.toUTFindex(pString, lResult);
    return lResult;
}

//-------------------------------------------------------
int ends(dchar[] pString, dchar[] pSubString)
//-------------------------------------------------------
{
    if (pString.length < pSubString.length)
        return -1;
    if (pSubString.length == 0)
        return -1;

    if (pString[pString.length-pSubString.length .. $] == pSubString)
        return pString.length-pSubString.length;
    return -1;
}


//-------------------------------------------------------
char[] enquote(char[] pString, char pTrigger = ' ', char[] pPrefix = `"`, char[] pSuffix = `"`)
//-------------------------------------------------------
{
    if ( (pString.length > 0) &&
         (find(pString, pTrigger) != -1)
       )
        return pPrefix ~ pString ~ pSuffix;

    return pString;
}

//-------------------------------------------------------
wchar[] enquote(wchar[] pString, wchar pTrigger = ' ', wchar[] pPrefix = `"`, wchar[] pSuffix = `"`)
//-------------------------------------------------------
{
    wchar[] lTrigger;
    lTrigger.length = 1;
    lTrigger[0] = pTrigger;
    if ( (pString.length > 0) &&
         (find(toUTF8(pString), toUTF8(lTrigger) ) != -1)
       )
        return pPrefix ~ pString ~ pSuffix;

    return pString;
}

//-------------------------------------------------------
dchar[] enquote(dchar[] pString, dchar pTrigger = ' ', dchar[] pPrefix = `"`, dchar[] pSuffix = `"`)
//-------------------------------------------------------
{
    dchar[] lTrigger;
    lTrigger.length = 1;
    lTrigger[0] = pTrigger;
    if ( (pString.length > 0) &&
         (find(toUTF8(pString), toUTF8(lTrigger)) != -1)
       )
        return pPrefix ~ pString ~ pSuffix;

    return pString;
}

unittest
{ // IsLike
     debug(2) writefln("str.IsLike.UT00");
     assert( IsLike( cast(dchar[])"foobar", cast(wchar[])"foo?*"));
     debug(2) writefln("str.IsLike.UT01");
     assert( IsLike( cast(char[])"foobar", cast(wchar[])"foo*"));
     debug(2) writefln("str.IsLike.UT02");
     assert( IsLike( cast(char[])"foobar", cast(wchar[])"*bar"));
     debug(2) writefln("str.IsLike.UT03");
     assert( !IsLike( cast(char[])"", cast(wchar[])"foo*"));
     debug(2) writefln("str.IsLike.UT04");
     assert( !IsLike( cast(char[])"", cast(wchar[])"*bar"));
     debug(2) writefln("str.IsLike.UT05");
     assert( !IsLike( cast(char[])"", cast(wchar[])"?"));
     debug(2) writefln("str.IsLike.UT06");
     assert( IsLike( cast(char[])"", cast(wchar[])"*"));
     debug(2) writefln("str.IsLike.UT07");
     assert( IsLike( cast(char[])"f", cast(wchar[])"?"));
     debug(2) writefln("str.IsLike.UT08");
     assert( IsLike( cast(char[])"f", cast(wchar[])"*"));
     debug(2) writefln("str.IsLike.UT09");
     assert( IsLike( cast(char[])"foo", cast(wchar[])"?oo"));
     debug(2) writefln("str.IsLike.UT10");
     assert( !IsLike( cast(char[])"foobar", cast(wchar[])"?oo"));
     debug(2) writefln("str.IsLike.UT11");
     assert( IsLike( cast(char[])"foobar", cast(wchar[])"?oo*"));
     debug(2) writefln("str.IsLike.UT12");
     assert( IsLike( cast(char[])"foobar", cast(wchar[])"*oo*b*"));
     debug(2) writefln("str.IsLike.UT13");
     assert( IsLike( cast(char[])"foobar", cast(wchar[])"*oo*ar"));
     debug(2) writefln("str.IsLike.UT14");
     assert( IsLike( cast(char[])"terrainformatica.com", cast(wchar[])"*.com"));
     debug(2) writefln("str.IsLike.UT15");
     assert( IsLike( cast(char[])"12abcdef", cast(wchar[])"*abc?e*"));
     debug(2) writefln("str.IsLike.UT16");
     assert( IsLike( cast(char[])"12abcdef", cast(wchar[])"**abc?e**"));
     debug(2) writefln("str.IsLike.UT17");
     assert( IsLike( cast(char[])"12abcdef", cast(wchar[])"*"));
     debug(2) writefln("str.IsLike.UT18");
     assert( IsLike( cast(char[])"12abcdef", cast(wchar[])"?*"));
     debug(2) writefln("str.IsLike.UT19");
     assert( !IsLike( cast(char[])"12abcdef", cast(wchar[])"*?"));
     debug(2) writefln("str.IsLike.UT20");
     assert( !IsLike( cast(char[])"12abcdef", cast(wchar[])"?*?"));
     debug(2) writefln("str.IsLike.UT21");
     assert( IsLike( cast(char[])"12abcdef", cast(wchar[])"*?*"));
     debug(2) writefln("str.IsLike.UT22");
     assert( IsLike( cast(char[])"12", cast(wchar[])"??"));
     debug(2) writefln("str.IsLike.UT23");
     assert( !IsLike( cast(char[])"123", cast(wchar[])"??"));
     debug(2) writefln("str.IsLike.UT24");
     assert( !IsLike( cast(char[])"12", cast(wchar[])"??3"));
     debug(2) writefln("str.IsLike.UT25");
     assert( !IsLike( cast(char[])"12", cast(wchar[])"???"));
     debug(2) writefln("str.IsLike.UT26");
     assert( IsLike( cast(char[])"", cast(wchar[])""));
     debug(2) writefln("str.IsLike.UT27");
     assert( IsLike( cast(char[])"abc", cast(wchar[])"abc"));
     debug(2) writefln("str.IsLike.UT28");
     assert( !IsLike( cast(char[])"abc", cast(wchar[])""));
     debug(2) writefln("str.IsLike.UT29");
     assert( IsLike( cast(char[])"abc*d", cast(wchar[])"abc\\*d"));
     debug(2) writefln("str.IsLike.UT30");
     assert( IsLike( cast(char[])"abc?d", cast(wchar[])"abc\\?d"));
     debug(2) writefln("str.IsLike.UT31");
     assert( IsLike( cast(char[])"abc\\d", cast(wchar[])"abc\\\\d"));
     debug(2) writefln("str.IsLike.UT32");
     assert( IsLike( cast(dchar[])"foobar", cast(wchar[])"foo???"));
     debug(2) writefln("str.IsLike.UT33");
     assert( ! IsLike( cast(dchar[])"foobar", cast(wchar[])"foo????"));
}


unittest
{  // begins
     debug(2) writefln("str.begins.UT01");
     assert( begins( cast(char[])"foobar", cast(char[])"o") == -1);
     debug(2) writefln("str.begins.UT02");
     assert( begins( cast(char[])"foobar", cast(char[])"f") == 1);
     debug(2) writefln("str.begins.UT03");
     assert( begins( cast(char[])"foobar", cast(char[])"fo") == 2);
     debug(2) writefln("str.begins.UT04");
     assert( begins( cast(char[])"foobar", cast(char[])"foo") == 3);
     debug(2) writefln("str.begins.UT05");
     assert( begins( cast(char[])"foobar", cast(char[])"foob") == 4);
     debug(2) writefln("str.begins.UT06");
     assert( begins( cast(char[])"foobar", cast(char[])"fooba") == 5);
     debug(2) writefln("str.begins.UT07");
     assert( begins( cast(char[])"foobar", cast(char[])"foobar") == 6);
     debug(2) writefln("str.begins.UT08");
     assert( begins( cast(char[])"foobar", cast(char[])"foobarx") == -1);
     debug(2) writefln("str.begins.UT09");
     assert( begins( cast(char[])"foobar", cast(char[])"oo") == -1);
     debug(2) writefln("str.begins.UT10");
     assert( begins( cast(char[])"foobar", cast(char[])"") == -1);
     debug(2) writefln("str.begins.UT11");
     assert( begins( cast(char[])"", cast(char[])"") == -1);
}

unittest
{  // ends
     debug(2) writefln("str.ends.UT01");
     assert( ends( cast(char[])"foobar", cast(char[])"a") == -1);
     debug(2) writefln("str.ends.UT02");
     assert( ends( cast(char[])"foobar", cast(char[])"r") == 5);
     debug(2) writefln("str.ends.UT03");
     assert( ends( cast(char[])"foobar", cast(char[])"ar") == 4);
     debug(2) writefln("str.ends.UT04");
     assert( ends( cast(char[])"foobar", cast(char[])"bar") == 3);
     debug(2) writefln("str.ends.UT05");
     assert( ends( cast(char[])"foobar", cast(char[])"obar") == 2);
     debug(2) writefln("str.ends.UT06");
     assert( ends( cast(char[])"foobar", cast(char[])"oobar") == 1);
     debug(2) writefln("str.ends.UT07");
     assert( ends( cast(char[])"foobar", cast(char[])"foobar") == 0);
     debug(2) writefln("str.ends.UT08");
     assert( ends( cast(char[])"foobar", cast(char[])"foobarx") == -1);
     debug(2) writefln("str.ends.UT09");
     assert( ends( cast(char[])"foobar", cast(char[])"oo") == -1);
     debug(2) writefln("str.ends.UT10");
     assert( ends( cast(char[])"foobar", cast(char[])"") == -1);
     debug(2) writefln("str.ends.UT11");
     assert( ends( cast(char[])"", cast(char[])"") == -1);
}

char[] Expand(char[] pOriginal, char[] pTokenList,
                char[] pLeading = "{", char[] pTrailing = "}")
{
    return std.utf.toUTF8( Expand (
                    std.utf.toUTF32(pOriginal),
                    std.utf.toUTF32(pTokenList),
                    std.utf.toUTF32(pLeading),
                    std.utf.toUTF32(pTrailing)
                    ) ) ;

}

dchar[] Expand(dchar[] pOriginal, dchar[] pTokenList,
                dchar[] pLeading = "{", dchar[] pTrailing = "}")
{
    int lPos;

    dchar[] lResult;
    dchar[][] lTokens;


    lResult = pOriginal.dup;

    lTokens = TokenizeLine(pTokenList,cast(dchar[])",","","");

    foreach(dchar [] lKV; lTokens)
    {
        dchar[][] lKeyValue;
        dchar[] lToken;
        dchar[] lReplacement;
        lKeyValue = TokenizeLine(lKV, cast(dchar[])"=","","");
        lToken = pLeading ~ lKeyValue[0] ~ pTrailing;
        lReplacement = lKeyValue[1];
        while( true)
        {
            lPos = find(toUTF8(lResult), toUTF8(lToken) );
            if (lPos == -1)
                break;
            lResult = lResult[0..lPos] ~
                      lReplacement ~
                      lResult[lPos + lToken.length .. $];
        }
    }

    // Now remove all unused tokens.
    while( (lPos = find(toUTF8(lResult), toUTF8(pLeading) )) != -1)
    {
        int lPos2;

        lPos2 = find(toUTF8(lResult[lPos+1 .. $]), toUTF8(pTrailing) );
        if (lPos2 == -1)
            break;

        lResult = lResult[0 .. lPos] ~
                  lResult[lPos2 + lPos + pLeading.length + pTrailing.length .. $];
    }

    return lResult;
}

dchar[] stripl(dchar[] s)
{
    uint i;

    for (i = 0; i < s.length; i++)
    {
	if (!std.ctype.isspace(s[i]))
	    break;
    }
    return s[i .. s.length];
}

dchar[] stripr(dchar[] s)
{
    uint i;

    for (i = s.length; i > 0; i--)
    {
	if (!std.ctype.isspace(s[i - 1]))
	    break;
    }
    return s[0 .. i];
}

dchar[] strip(dchar[] s)
{
    return stripr(stripl(s));
}

unittest
{
    dchar[] s;

    s = strip(cast(dchar[])"  foo\t ");
    assert(s == "foo");
    s = stripr(cast(dchar[])"  foo\t ");
    assert(s == "  foo");
    s = stripl(cast(dchar[])"  foo\t ");
    assert(s == "foo\t ");
}

char[] toASCII(char[] pUTF8)
{
    bool lChanged;
    char[] lResult;
    // Convert non-ASCII chars based on the Microsoft DOS Western Europe charset
    static char[] lTranslateTable =
        "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0A\x0B\x0C\x0D\x0E\x0F"
        "\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1A\x1B\x1C\x1D\x1E\x1F"
        " !\"#$%&'()*+,-./0123456789:;<=>?"
        "@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_"
        "`abcdefghijklmnopqrstuvwxyz{|}~ "
        "CueaaaaceeeiiiAAEaAooouuyOUo$Oxf"
        "aiounNa0?R!24!<>-----AAAC----c$-"
        "------aA-------$dDEEEiIII----|I-"
        "OBOOoOmdDUUUyY-'-+=3PS/,0:.132- ";
    lResult = pUTF8;
    for (int i = 0; i < lResult.length; i++)
    {
        if (lResult[i] > 127)
        {
            if (lChanged == false)
            {
                lResult = pUTF8.dup;
                lChanged = true;
            }
            lResult[i] = lTranslateTable[lResult[i]];
        }
    }
    return lResult;
}